(def i18n-version "1.0.4")
(def slf4j-version "2.0.17")

(defproject org.openvoxproject/clj-shell-utils "2.2.0"
  :description "Clojure shell execution utilities"

  :min-lein-version "2.9.1"

  :license {:name "Apache-2.0"
            :url "https://www.apache.org/licenses/LICENSE-2.0.txt"}
            
  :pedantic? :abort

  :test-paths ["test/unit"]

  :plugins [[lein-project-version "0.1.0"]
            [org.openvoxproject/i18n ~i18n-version]]

  :source-paths ["src/clj"]
  :java-source-paths ["src/java"]

  ;; Generally, try to keep version pins in :managed-dependencies and the libraries
  ;; this project actually uses in :dependencies, inheriting the version from
  ;; :managed-dependencies. This prevents endless version conflicts due to deps of deps.
  ;; Renovate should keep the versions largely in sync between projects.
  :managed-dependencies [[org.clojure/clojure "1.12.4"]
                         [org.clojure/tools.logging "1.3.1"]

                         [clj-time "0.15.2"]
                         [commons-io "2.22.0"]
                         [org.apache.commons/commons-exec "1.6.0"]
                         [org.slf4j/log4j-over-slf4j ~slf4j-version]
                         [org.slf4j/slf4j-api ~slf4j-version]
                         [org.openvoxproject/i18n ~i18n-version]
                         [org.openvoxproject/kitchensink "3.5.7"]
                         [org.openvoxproject/trapperkeeper "5.0.0"]
                         [org.openvoxproject/trapperkeeper "5.0.0" :classifier "test"]
                         [prismatic/schema "1.4.1"]]

  :dependencies [[org.clojure/clojure]
                 [commons-io]
                 [org.apache.commons/commons-exec]
                 [org.openvoxproject/i18n]
                 [org.openvoxproject/kitchensink]
                 [org.openvoxproject/trapperkeeper]
                 [org.slf4j/log4j-over-slf4j]
                 [org.slf4j/slf4j-api]
                 [prismatic/schema]]



  :profiles { :test { :dependencies [[org.openvoxproject/trapperkeeper :classifier "test" :scope "test"]]}}

  :deploy-repositories [["releases" {:url "https://clojars.org/repo"
                                     :username :env/CLOJARS_USERNAME
                                     :password :env/CLOJARS_PASSWORD
                                     :sign-releases false}]])
